#ifndef ARCLIB_REGISTRAR
#define ARCLIB_REGISTRAR

#include <string>

#include <arc/error.h>


/**
 * Error registrar class.
 */
class RegistrarError : public ARCLibError {

    public:
        /** Constructor for the RegistrarError exception. Calls the 
		 * corresponding constructor in ARCLibError.
         */
        RegistrarError(std::string message) : ARCLibError(message) {};
};


/**
 * This class provides interface for registering plug-in functionality.
 * Classes which implement functionality may inherit from this class to
 * type casting cleaner.
 * Usage example 
 * In external library:
 *   class NewFunctionality: public Functionality {
 *     ...
 *   };
 *
 *   NewFunctionality& GetNewFunctionality(const std::string& id) {
 *     if(!CheckID(id)) throw ARCLibError("");
 *     return *(new NewFunctionality(id));
 *   }
 *
 *   static RegistrarItem("functionality",&GetNewFunctionality);
 *
 * In main code:
 *   class Functionality: private RegistrarItem {
 *     ...
 *   };
 *
 *   Functionality& func = RegistrarItem::GetItem("functionality","identifier");
 *
 */
class RegistrarItem {
	public:
		/**
		 * Classes implementing new functionality should have 
		 * corresponding function constructor of this type.
		 */
		typedef RegistrarItem& (*GetItem_t)(const std::string& id);
		/**
		 * Dummy constructor to be used in inherited classes.
		 */
		RegistrarItem(void);
		/**
		 * This contructor should be used to register new functionality.
		 * Simplest way to do that is to define single static element.
		 */
		RegistrarItem(const std::string& name,GetItem_t get_item);
		/**
		 * Returns class implementing requested functionality or 
		 * throws exception.
		 */
		static RegistrarItem& GetItem(const std::string& name,const std::string& id);
};

#endif
