package se.umu.jobsubmission.client; 

import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.net.InetAddress;
import java.util.LinkedList;
import java.util.List;
import java.util.Vector;

import javax.xml.rpc.Stub;

import org.apache.axis.AxisFault;
import org.apache.axis.message.addressing.Address;
import org.apache.axis.message.addressing.EndpointReferenceType;
import org.ggf.schemas.jsdl._2005._11.jsdl.JobDefinition_Type;
import org.globus.axis.util.Util;
import org.globus.wsrf.client.ServiceURL;
import org.globus.wsrf.config.ContainerConfig;
import org.globus.wsrf.encoding.DeserializationException;
import org.globus.wsrf.encoding.ObjectDeserializer;
import org.globus.wsrf.impl.security.authentication.Constants;
import org.globus.wsrf.utils.FaultHelper;
import org.oasis.wsrf.faults.BaseFaultType;
import org.xml.sax.InputSource;

import se.umu.jobsubmission.ClientConfiguration;
import se.umu.jobsubmission.InvalidJobDescriptionFaultType;
import se.umu.jobsubmission.JobPreferencesType;
import se.umu.jobsubmission.JobRequest;
import se.umu.jobsubmission.CoallocationRequest;
import se.umu.jobsubmission.JobRequestFaultType;
import se.umu.jobsubmission.JobSubmission;
import se.umu.jobsubmission.JobSubmissionServiceAddressingLocator;
import se.umu.jobsubmission.NoAptResourcesFoundFaultType;
import se.umu.jobsubmission.NoResourcesFoundFaultType;
import se.umu.jobsubmission.CoallocateJobResponse;
import se.umu.jobsubmission.impl.jobsubmissionservice.JobSubmissionConstants;
import se.umu.jobsubmission.util.ClientUtils;
import se.umu.jobsubmission.util.JobDescriptionTranslator;


/**
 * Simple client for job submission. 
 * The client accepts four parameters, 
 * the job description (mandatory),
 * a job preferences document (optional), 
 * a list of index servers (optional), 
 * and an URL to the JSS to use (optional)
 */

public class CoallocationClient {

    static {
        Util.registerTransport(); 
    }

    private List jobDescriptionFiles; 
    private List jobPreferencesFiles;
    private String indexServers[]; 
    private String URI; 
    private String host; 
    private String clientHost; 
    private boolean clientStagesInput; 
    private Integer tolerance; 
    
    private String[] converterClasses; 
    private String[] stagingTargetResolverClasses; 
    
    private JobDescriptionTranslator jobDescriptionTranslators[]; 

    private static final String optionsDesc = 
        "OPTIONS:\n" +
        "\n"+
        "Help:\n" +
        " -help                 display help.\n" +
        " -usage                display usage.\n" +
        "Usage:\n" + 
        " -f                    Specify file that contains job description\n" + 
        " -i                    Specify IndexService to retrieve resources from\n" + 
        " -j                    Specify URL of job submission service to contact\n" +
        " -p                    Specify file that contains job preferences\n" +
		" -t                    Specify allowed job start time deviation\n" + 
		" -s                    Disable client file staging\n" + 
        "\n"; 

    private static final String usageDesc =
        "ARGUMENTS: \n" +
        "         [options] -f <job description file name>\n" +
        "with\n" +
        "         [options]       [-i <indexService]+ \n" +
        "                         [-p <job preferences file name>]\n" +
        "                         [-j <job submission service> (default is localhost)]\n" +
		"                         [-t <job tolerance (in seconds)>\n" +
		"                         [-s]\n" + 
        "\n";

    public CoallocationClient() throws Exception {
    	
        jobDescriptionFiles = new LinkedList(); 
        jobPreferencesFiles = new LinkedList();  
        // Using resolved address instead of 127.0.0.1 which gives 
        // some problems with authentication as certificates differs. 
        InetAddress address = InetAddress.getLocalHost();     
        this.host = address.getCanonicalHostName(); 
        this.clientHost = host; 
        
        this.URI = ServiceURL.getURL(
        		host, JobSubmissionConstants.SERVICE_NAME).
					toString(); 
        
        this.indexServers = null; 
        
        this.clientStagesInput = true; 
        
        this.tolerance = null; 
        
        // Read and deserialize config file 
    	String configFile = "/etc/jobsubmission/clientconfig.xml"; 
    	String globusLocation = ContainerConfig.getGlobusLocation(); 
    	String filePath = globusLocation + configFile; 
        
        ClientConfiguration config = null;
        
        FileInputStream input = new FileInputStream(filePath); 
        InputSource inputSource = new InputSource(input); 
        config = (ClientConfiguration) ObjectDeserializer.deserialize(inputSource, 
        		ClientConfiguration.class); 
        
        this.converterClasses = config.getDescriptionConverter(); 
        this.stagingTargetResolverClasses = config.getStagingTargetResolver(); 
        
    }
      
    /**
     * Parse user input
     * @param args
     * @throws Exception
     */
    public void processArguments(String args[]) throws Exception {

        List indexServersList = new Vector(); 

        if(args.length == 0) {
            System.out.println();
            System.out.println(usageDesc);
            System.out.println(optionsDesc);
            System.exit(0); 
        }
        
        int noJobDescs = 0; 
        int noJobPrefs = 0; 

        boolean error = false; 
        for(int i = 0; i < args.length && !error; i++) {

            if (args[i].startsWith("-")) {

                if (args[i].equalsIgnoreCase("-help")) {
                    System.out.println();
                    System.out.println(usageDesc);
                    System.out.println(optionsDesc);
                    System.exit(0);
                }
                else if (args[i].equalsIgnoreCase("-usage")) {
                    System.out.println(usageDesc);
                    System.exit(0);
                }
                else if (args[i].equals("-f")) {
                    i++; 
                    if(i == args.length) {
                        System.err.println("-f requires a job description file");
                        error = true; 
                    }
                    else
                        jobDescriptionFiles.add(args[i]); 
                }
                else if (args[i].equals("-i")) {
                    i++; 
                    if(i == args.length) {
                        System.err.println("-i requires a index service URI");
                        error = true; 
                    }
                    else 
                        indexServersList.add(args[i]); 
                }
                else if (args[i].equals("-j")) {
                    i++; 
                    if(i == args.length) {
                        System.err.println("-j requires a job submission service URI");
                        error = true; 
                    }
                    else  {
                        host = args[i]; 
                
                        URI = ServiceURL.getURL(
                        		host, 
								JobSubmissionConstants.SERVICE_NAME).toString();
                    }
							
                }
                else if (args[i].equals("-p")) {
                    i++; 
                    if(i == args.length) {
                        System.err.println("-p requires a job preferences file");
                        error = true; 
                    }
                    else {
                        jobPreferencesFiles.add(args[i]); 
                    }
                }
                else if (args[i].equals("-s")) {
                	System.out.println("File staging will be performed by "
                			+ "service, not by client"); 
                	this.clientStagesInput = false; 
                }
                else if (args[i].equals("-t")) {
                	i++; 
                	if(i == args.length) {
                		System.err.println("-t requires a tolerance"); 
                		error = true; 
                	}
                	else {
                		try { 
                			this.tolerance = new Integer(args[i]); 
                			if(this.tolerance.intValue() < 0) {
                				System.err.println("tolerance must be positive"); 
                				error = true; 
                			}
                		} catch (Exception e) {
                			System.err.println("Invalid tolerance " + args[i]); 
                			error = true; 
                		}
                	}
                }
                else {
                    System.err.println("Unknown argument: " + args[i]);
                    error = true; 
                }
            }
            else {
                System.err.println("Unknown argument: " + args[i]);
                error = true; 
            }
        }
        if(error) {
            System.exit(1); 
        }

        int noIndexServers = indexServersList.size(); 

        if(noIndexServers > 0) {
            indexServers = new String[noIndexServers]; 
            for(int i = 0; i < noIndexServers; i++) 
                indexServers[i] = (String) indexServersList.get(i); 
        }
    }
       
     
    /**
     * Locate JobSubmissionService, convert job description
     * and submit it to service. Include any user preferences document
     * and/or non-default index server(s) in the job request. 
     * @throws Exception
     */
    public void coallocateJob() throws FileNotFoundException, Exception {

        System.out.println("Submitting job using:\n" + URI);
        
        // Perform some basic input checks. 
        System.out.println("Converting job description to JSDL");
        if(jobDescriptionFiles.size() == 0) {
        	System.err.println("The -f flag (JobDescription) must be specified"); 
        	throw new 
				Exception("The -f flag (JobDescription) must be specified"); 
        }        
        
        if(jobDescriptionFiles.size() < 2) {
        	System.err.println("At least 2 job description files must be " 
        			+ "specified"); 
        	throw new Exception("At least 2 job description files must be "
        			+ "specified"); 
        }
        
        if(jobPreferencesFiles.size() > jobDescriptionFiles.size()) {
        	System.err.println("Too many job preference files specified"); 
        	throw new Exception("Too many job preference files specified"); 
        }
        
        // Convert job descriptions.
        // If one of these conversion fails, the whole coallocation attempt fails. 
        JobDefinition_Type jobDescs[] = new JobDefinition_Type[jobDescriptionFiles.size()]; 
        for(int i = 0; i < jobDescriptionFiles.size(); i++) {
        	jobDescs[i] = ClientUtils.convertJobDescription(
        			(String)jobDescriptionFiles.get(i), 
					this.converterClasses); 
        }
        
        // Jobpreferences file 
        // User may specify 0 or more job preference files.         
        JobPreferencesType preferences[] = new JobPreferencesType[jobPreferencesFiles.size()]; 
        for(int i = 0; i < jobPreferencesFiles.size(); i++) {
        	JobPreferencesType tmp =  ClientUtils.getJobPreferences(
        			(String)jobPreferencesFiles.get(i)); 
        	if(tmp != null) {
        		preferences[i] = tmp; 
        	} 
        	else {
        		System.err.println("Incorrect job preferences file: " 
        				+ jobPreferencesFiles.get(i)); 
        		throw new Exception("Incorrect job preferences file: " 
        				+ jobPreferencesFiles.get(i)); 
        	}
        }
        

        // Locate service
        System.out.println("Locating job submission service"); 
        JobSubmissionServiceAddressingLocator locator = 
            new JobSubmissionServiceAddressingLocator(); 
        JobSubmission jobsubmissionservice = null; 
        try { 
        	EndpointReferenceType endpoint = new EndpointReferenceType();
        	endpoint.setAddress(new Address(URI));
        	jobsubmissionservice = locator.getJobSubmissionPort(endpoint);        
        } catch (Exception e) {
        	System.err.println("Can not contact job submission service at " + URI); 
        	throw e; 
        }
        
        // Client security descriptor file 
        String clientDesc = "etc/jobsubmission/client-security-config.xml"; 
        // Set descriptor on Stub 
        ((Stub)jobsubmissionservice)._setProperty(
        		Constants.CLIENT_DESCRIPTOR_FILE, 
				clientDesc);
                 
        // Build coallocation request message
        CoallocationRequest requestParameters = new CoallocationRequest(); 
        requestParameters.setIndexServerURIs(this.indexServers); 
        requestParameters.setClientURL(this.clientHost); 
        requestParameters.setClientStagesFiles(new Boolean(this.clientStagesInput)); 
        requestParameters.setTolerance(tolerance); 
        
        // Add job descriptions
        requestParameters.setJobDescription(jobDescs); 
        // Add job preferences files: 
        requestParameters.setJobPreferences(preferences); 
        
        CoallocateJobResponse response = null; 
 
        System.out.println("Coallocating job..."); 
        
        try { 
        	response = jobsubmissionservice.coallocateJob(requestParameters); 
        }  
        catch (InvalidJobDescriptionFaultType f) {
        	System.err.println("Job request failed, invalid job description:"); 
        	FaultHelper helper = new FaultHelper(f); 
        	System.err.println(helper.getDescriptionAsString()); 
        } catch(NoResourcesFoundFaultType f) {
        	System.err.println("Job request failed"); 
        	FaultHelper helper = new FaultHelper(f); 
        	System.err.println(helper.getDescriptionAsString()); 
        } catch(NoAptResourcesFoundFaultType f) {
        	System.err.println("Job request failed"); 
        	FaultHelper helper = new FaultHelper(f); 
        	System.out.println(helper.getDescriptionAsString()); 
        } catch(JobRequestFaultType f) {
        	System.err.println("Job request failed:"); 
        	FaultHelper helper = new FaultHelper(f); 
        	System.err.println(helper.getDescriptionAsString()); 
        } catch (BaseFaultType f) {        	
        	System.err.println("General fault: "); 
        	FaultHelper helper = new FaultHelper(f);         	        	        	
        	System.err.println(helper.getDescriptionAsString());         	
        } catch(AxisFault f) {
        	f.printStackTrace(); 
        	System.err.println("Can not contact job submission service: " + URI); 
        	System.err.println("Reason: " + f.getFaultReason()); 
        } 
        
        if(response != null) {
        	int i = 0; 
        	if(this.clientStagesInput) {
        		try { 
        			for(i = 0; i < jobDescs.length; i++) {
        				
        				System.out.println("Staging input files for job " + i); 
        				if(ClientUtils.stageJobInput(response.getJobIds(i), 
        						jobDescs[i], 
								this.stagingTargetResolverClasses, 
								this.clientHost)) {
        					System.out.println("... done"); 
        				}
        				else {
        					System.out.println("... no input files to stage"); 
        				}
        			}
        		} catch (Exception e) {
        			System.err.println("Error staging input files: " + e.getMessage());
        			System.err.println("Cancelling all jobs"); 
        			for(int j = 0; j < i; j++) {
        				System.out.println("cancelling job " + response.getJobIds(j)); 
        				ClientUtils.cancelJob(response.getJobIds(j), 
        						this.stagingTargetResolverClasses); 
        			}
        			throw e; 
        		}
        	}
        	
        	String epr = response.getEndpointReference().toString(); 
        	System.out.println("Epr to all subjobs: " + epr.toString()); 
        	String jobIds[] = response.getJobIds(); 
        	if(jobIds != null) {
        		System.out.println("Job submitted with jobIds: "); 
        		for(i = 0; i < jobIds.length; i++) {
        			System.out.println(jobIds[i].toString()); 
        		}
        	}
        	        
        } 
    }
   

    public static void main(String args[]) {

        try {

            CoallocationClient client = new CoallocationClient(); 
            client.processArguments(args); 
            client.coallocateJob(); 

        } catch (FileNotFoundException e) {
        	System.err.println(e.getMessage()); 
        }
        catch (Exception e) {
            System.out.println("Error submitting job");
        }
    }
}


